/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_main.c

#include "quakedef.h"

entity_t	r_worldentity;

vec3_t		modelorg, r_entorigin;
entity_t	*currententity;

int			r_visframecount;	// bumped when going to a new PVS
int			r_framecount;		// used for dlight push checking

mplane_t	frustum[4];

int			c_brush_polys, c_alias_polys;

int			particletexture;	// little dot for particles
int			playertextures;		// up to 16 color translated skins

//
// view origin
//
vec3_t	vup;
vec3_t	vpn;
vec3_t	vright;
vec3_t	r_origin;

//
// screen size info
//
refdef_t	r_refdef;

mleaf_t		*r_viewleaf, *r_oldviewleaf;

int		d_lightstylevalue[256];	// 8.8 fraction of base light value


void R_MarkLeaves (void);
void R_RenderSky (void);

cvar_t	r_norefresh = {"r_norefresh","0"};
cvar_t	r_drawentities = {"r_drawentities","1"};
cvar_t	r_drawviewmodel = {"r_drawviewmodel","1"};
cvar_t	r_speeds = {"r_speeds","0"};
cvar_t	r_novis = {"r_novis","0"};

cvar_t	gl_cull = {"gl_cull","1"};
cvar_t	gl_polyblend = {"gl_polyblend","1"};
cvar_t	gl_nocolors = {"gl_nocolors","0"};
cvar_t	gl_doubleeyes = {"gl_doubleeys", "1"};

void R_DrawAliasModel (entity_t *e);
void R_DrawBrushModel (entity_t *e);
void R_TextureFrame (void);

// for model drawing
extern polyverts_t *AliasVerts;
extern polyverts_t *polyverts;

// occlusion query objects
int OcQueries[MAX_VISEDICTS];

typedef struct gib_s
{
	vec3_t origin;
	vec3_t angles;
	model_t *model;
	mleaf_t *leaf;
} gib_t;

gib_t r_giblist[16384];
int numgibsonlist = 0;

void R_AddGibToList (entity_t *ent)
{
	gib_t *gib;
	
	// no more gib slots
	if (numgibsonlist >= 16384)
		return;

	// take a gib slot
	gib = &r_giblist[numgibsonlist];
	numgibsonlist++;

	// copy in the data
	gib->angles[0] = ent->lastgoodangles[0];
	gib->angles[1] = ent->lastgoodangles[1];
	gib->angles[2] = ent->lastgoodangles[2];

	gib->origin[0] = ent->lastgoodorigin[0];
	gib->origin[1] = ent->lastgoodorigin[1];
	gib->origin[2] = ent->lastgoodorigin[2];

	gib->model = ent->model;

	// get the leaf for visibility checking
	gib->leaf = Mod_PointInLeaf (gib->origin, cl.worldmodel);
}


/*
=================
R_CullBox

Returns true if the box is completely outside the frustom
=================
*/
qboolean R_CullBox (vec3_t mins, vec3_t maxs)
{
	int		i;

	for (i=0 ; i<4 ; i++)
		if (BoxOnPlaneSide (mins, maxs, &frustum[i]) == 2)
			return true;
	return false;
}


/*
=============================================================

  SPRITE MODELS

=============================================================
*/

/*
================
R_GetSpriteFrame
================
*/
mspriteframe_t *R_GetSpriteFrame (entity_t *currententity)
{
	msprite_t		*psprite;
	mspritegroup_t	*pspritegroup;
	mspriteframe_t	*pspriteframe;
	int				i, numframes, frame;
	float			*pintervals, fullinterval, targettime, time;

	psprite = currententity->model->cache.data;
	frame = currententity->frame;

	if ((frame >= psprite->numframes) || (frame < 0))
	{
		Con_Printf ("R_DrawSprite: no such frame %d\n", frame);
		frame = 0;
	}

	if (psprite->frames[frame].type == SPR_SINGLE)
	{
		pspriteframe = psprite->frames[frame].frameptr;
	}
	else
	{
		pspritegroup = (mspritegroup_t *)psprite->frames[frame].frameptr;
		pintervals = pspritegroup->intervals;
		numframes = pspritegroup->numframes;
		fullinterval = pintervals[numframes-1];

		time = cl.time + currententity->syncbase;

	// when loading in Mod_LoadSpriteGroup, we guaranteed all interval values
	// are positive, so we don't have to worry about division by 0
		targettime = time - ((int)(time / fullinterval)) * fullinterval;

		for (i=0 ; i<(numframes-1) ; i++)
		{
			if (pintervals[i] > targettime)
				break;
		}

		pspriteframe = pspritegroup->frames[i];
	}

	return pspriteframe;
}


/*
=================
R_DrawSpriteModel

=================
*/
void R_DrawSpriteModel (entity_t *e)
{
	vec3_t	point;
	mspriteframe_t	*frame;
	float		*up, *right;
	vec3_t		v_forward, v_right, v_up;
	msprite_t		*psprite;

	// don't even bother culling, because it's just a single
	// polygon without a surface cache
	frame = R_GetSpriteFrame (e);
	psprite = currententity->model->cache.data;

	if (psprite->type == SPR_ORIENTED)
	{
		// bullet marks on walls
		AngleVectors (currententity->angles, v_forward, v_right, v_up);
		up = v_up;
		right = v_right;
	}
	else
	{	// normal sprite
		up = vup;
		right = vright;
	}

	glColor3f (1, 1, 1);

	glBindTexture (GL_TEXTURE_2D, frame->glt->texnum);

	glBegin (GL_QUADS);

	glTexCoord2f (0, 1);
	VectorMA (e->origin, frame->down, up, point);
	VectorMA (point, frame->left, right, point);
	glVertex3fv (point);

	glTexCoord2f (0, 0);
	VectorMA (e->origin, frame->up, up, point);
	VectorMA (point, frame->left, right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 0);
	VectorMA (e->origin, frame->up, up, point);
	VectorMA (point, frame->right, right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 1);
	VectorMA (e->origin, frame->down, up, point);
	VectorMA (point, frame->right, right, point);
	glVertex3fv (point);
	
	glEnd ();
}

//==================================================================================


// hw version of the below using occlusion queries
// [quote="me"]unless we ever do something sexy with particles[/quote]
// well, here it is, then...
void R_HardwareOcclusion (void)
{
	int i;
	entity_t *ent;
	int numOccluded = 0;

	float mins[3];
	float maxs[3];

	byte boxindexes[] =
	{
		0, 1, 2, 3,
		4, 5, 1, 0,
		3, 2, 6, 7,
		5, 4, 7, 6,
		1, 5, 6, 2,
		4, 0, 3, 7
	};

	float boxverts[24];

	// disable texturing
	glDisable (GL_TEXTURE_2D);

	// because we don;t know the orientation of the bbox in advance...
	glDisable (GL_CULL_FACE);

	// disable framebuffer and depthbuffer writes
	glColorMask (GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	glDepthMask (GL_FALSE);

	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (3, GL_FLOAT, 0, boxverts);

	for (i = 0; i < cl_numvisedicts; i++)
	{
		ent = cl_visedicts[i];

		// don;t bother with these are they're included with the world
		if (ent->model->name[0] == '*') continue;

		// get mins and maxs points
		VectorAdd (ent->origin, ent->model->mins, mins);
		VectorAdd (ent->origin, ent->model->maxs, maxs);

		// CPU grunt to the rescue!!!
		boxverts[0] = boxverts[9] = boxverts[12] = boxverts[21] = mins[0];
		boxverts[3] = boxverts[6] = boxverts[15] = boxverts[18] = maxs[0];
		boxverts[1] = boxverts[4] = boxverts[13] = boxverts[16] = maxs[1];
		boxverts[7] = boxverts[10] = boxverts[19] = boxverts[22] = mins[1];
		boxverts[2] = boxverts[5] = boxverts[8] = boxverts[11] = maxs[2];
		boxverts[14] = boxverts[17] = boxverts[20] = boxverts[23] = mins[2];

		// begin the occlusion query
		glBeginQuery (GL_SAMPLES_PASSED, OcQueries[i]);

		// draw as indexed varray
		glDrawElements (GL_QUADS, 24, GL_UNSIGNED_BYTE, boxindexes);

		// end the query
		// don;t read back immediately so that we give the query time to be ready
		glEndQuery (GL_SAMPLES_PASSED);
	}

	glDisableClientState (GL_VERTEX_ARRAY);

	// restore basic state
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_CULL_FACE);

	// enable framebuffer and depthbuffer writes
	glColorMask (GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	glDepthMask (GL_TRUE);

	// some implementations don't reset the primary colour properly after restoring the colormask
	glColor4f (1, 1, 1, 1);

	// now we read back
	for (i = 0; i < cl_numvisedicts; i++)
	{
		ent = cl_visedicts[i];

		// don;t bother with these are they're included with the world
		if (ent->model->name[0] == '*') continue;

		// get the object and store it in the occlusion bits for the ent
		glGetQueryObjectiv (OcQueries[i], GL_QUERY_RESULT, &ent->OcclusionBits);

		if (!ent->OcclusionBits) numOccluded++;
	}

	// testing
	// Con_Printf ("Occluded %i Entities\n", numOccluded);
}

void TraceLine (vec3_t start, vec3_t end, vec3_t impact);

qboolean TraceOcclude (vec3_t org)
{
	vec3_t stop;
	float dist;

	TraceLine (org, r_refdef.vieworg, stop);
	dist = DotProduct (r_refdef.vieworg, stop);

	if (dist < 1) return false;

	return true;
}

qboolean R_ModelOcclusion (entity_t *ent)
{
	vec3_t org;
	vec3_t mins, maxs;

	// already tested and passed
	if (GL_Caps.GL_QueryBits) return false;

	// if any of the occlusion tests pass we return false and draw the flare
	// 1 - the origin
	// most flares will pass at this stage and return false
	if (!TraceOcclude (ent->origin)) return false;

	// get mins and maxs points
	VectorAdd (ent->origin, ent->model->mins, mins);
	VectorAdd (ent->origin, ent->model->maxs, maxs);

	// 2 - mins and maxs points
	if (!TraceOcclude (mins)) return false;
	if (!TraceOcclude (maxs)) return false;

	return true;
}

// particle effects
void R_FlareSprite (vec3_t org, int colour, float radius, float scale, float alpha);
void R_SLightSprite (vec3_t org);
void R_BubbleSprite (vec3_t org);

void R_AddEntModelEffects (entity_t *ent)
{
	vec3_t baseorigin = {ent->origin[0], ent->origin[1], ent->origin[2]};
	float baseradius = 25.0f;
	int flarecolour = 0;
	float scale = 0.0f;
	float alpha = 0.5f;

	// no effects
	if (!ent->model->effects) return;

	// handle sprite replacements first
	if (ent->model->effects & EF_S_LIGHT)
	{
		R_SLightSprite (ent->origin);
		return;
	}

	if (ent->model->effects & EF_BUBBLESPRITE)
	{
		R_BubbleSprite (ent->origin);
		return;
	}

	// because we can stack these effects we check them individually rather than making a mess
	if (ent->model->effects & EF_FIREFLARE)
	{
		flarecolour = 107;
		scale = 0.008f;
	}

	if (ent->model->effects & EF_TORCHFLARE)
	{
		baseorigin[2] += 8.0f;
		flarecolour = 107;
		scale = 0.008f;
	}

	if (ent->model->effects & EF_ROCKETFLARE)
	{
		// move the glow behind the rocket - fixme - do a fast sin/cos lookup table for this...
		baseorigin[0] += cos (ent->angles[1] * DIV180MULTPI) * (-20.0f);
		baseorigin[1] += sin (ent->angles[1] * DIV180MULTPI) * (-20.0f);
		baseorigin[2] += sin (ent->angles[0] * DIV180MULTPI) * (-20.0f);

		flarecolour = 107;
		alpha = 0.75f;
	}

	if (ent->model->effects & EF_FLICKER) baseradius += (rand () % (int) (baseradius / 6));

	// catch anything that hasn't had a colour set
	if (!flarecolour) return;

	// test for occlusion on the bbox
	if (R_ModelOcclusion (ent)) return;

	// add the flare
	R_FlareSprite (baseorigin, flarecolour, baseradius, scale, alpha);
}


/*
=============
R_DrawEntitiesOnList
=============
*/
void R_DrawEntitiesOnList (void)
{
	int		i;
	qboolean haveSprites = false;
	qboolean haveAlias = false;
	qboolean haveBrush = false;
	entity_t gibent;

	// don;t draw entities
	if (!r_drawentities.value) return;

	// hardware occlusion if we have it
	if (GL_Caps.GL_QueryBits) R_HardwareOcclusion ();

	// do a run through the entities to see what we have
	// even though we detect sprite models we don't bother drawing them, cos they're crap
	for (i = 0; i < cl_numvisedicts; i++)
	{
		// these are drawn with the world model
		if (cl_visedicts[i]->model->name[0] == '*') continue;

		// occluded
		if (!cl_visedicts[i]->OcclusionBits) continue;

		if (cl_visedicts[i]->model->type == mod_alias)
			haveAlias = true;
		else if (cl_visedicts[i]->model->type == mod_sprite)
			haveSprites = true;
		else haveBrush = true;

		// deal with entity model effects
		R_AddEntModelEffects (cl_visedicts[i]);
	}

	// alias models only if we have them
	if (haveAlias || numgibsonlist)
	{
		// clear the giblets temp entity
		memset (&gibent, 0, sizeof (entity_t));

		// set global rendering state for alias models
		glShadeModel (GL_SMOOTH);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

		// bring up multitexturing
		glActiveTexture (GL_TEXTURE1);
		glEnable (GL_TEXTURE_2D);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_ADD);

		GL_EnableVertexArrays (AliasVerts, VA_STDUAL | VA_RGBA);

		for (i = 0; i < cl_numvisedicts; i++)
		{
			// occluded
			if (!cl_visedicts[i]->OcclusionBits) continue;

			currententity = cl_visedicts[i];

			if (currententity->model->type == mod_alias) R_DrawAliasModel (currententity);
		}

		for (i = 0; i < numgibsonlist; i++)
		{
			// not visible in this visframe
//			if (r_giblist[i].leaf->visframe != r_visframecount)
//				continue;

//			Con_Printf ("Drawing from gib list\n");

			// set current entity
			currententity = &gibent;

			// copy the gib across
			currententity->angles[0] = r_giblist[i].angles[0];
			currententity->angles[1] = r_giblist[i].angles[1];
			currententity->angles[2] = r_giblist[i].angles[2];

			currententity->origin[0] = r_giblist[i].origin[0];
			currententity->origin[1] = r_giblist[i].origin[1];
			currententity->origin[2] = r_giblist[i].origin[2];

			currententity->model = r_giblist[i].model;

			// draw it
			R_DrawAliasModel (currententity);
		}

		// restore state
		GL_DisableVertexArrays ();

		glDisable (GL_TEXTURE_2D);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glActiveTexture (GL_TEXTURE0);

		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glShadeModel (GL_FLAT);
		glColor3f (1, 1, 1);
	}

	// boxes only if we have them
	if (haveBrush)
	{
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

		// bring up multitexturing
		glActiveTexture (GL_TEXTURE1);
		glEnable (GL_TEXTURE_2D);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_ADD);

		GL_EnableVertexArrays (polyverts, VA_STDUAL);

		for (i = 0; i < cl_numvisedicts; i++)
		{
			// occluded
			if (!cl_visedicts[i]->OcclusionBits) continue;

			currententity = cl_visedicts[i];

			if (currententity->model->type == mod_brush)
			{
				// these are drawn with the world model
				if (cl_visedicts[i]->model->name[0] == '*')
					continue;

				// it's a box model so draw it
				R_DrawBrushModel (currententity);
			}
		}

		// restore state
		GL_DisableVertexArrays ();

		glDisable (GL_TEXTURE_2D);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glActiveTexture (GL_TEXTURE0);

		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glColor3f (1, 1, 1);
	}
}

/*
=============
R_DrawViewModel
=============
*/
void R_DrawViewModel (void)
{
	if (!r_drawviewmodel.value) return;
	if (chase_active.value) return;
	if (!r_drawentities.value) return;
	if (cl.stats[STAT_HEALTH] <= 0) return;

	currententity = &cl.viewent;

	if (!currententity->model) return;

	// hack the depth range to prevent view model from poking into walls
	glDepthRange (0.1, 0.2);

	// set state seperately here again
	glShadeModel (GL_SMOOTH);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	// bring up multitexturing
	glActiveTexture (GL_TEXTURE1);
	glEnable (GL_TEXTURE_2D);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_ADD);

	GL_EnableVertexArrays (AliasVerts, VA_STDUAL | VA_RGBA);

	// draw it
	R_DrawAliasModel (currententity);

	// restore state
	GL_DisableVertexArrays ();

	glDisable (GL_TEXTURE_2D);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glActiveTexture (GL_TEXTURE0);

	// restore state
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glShadeModel (GL_FLAT);
	glColor3f (1, 1, 1);

	// restore depth range
	glDepthRange (0.3, 1);
}


/*
============
R_PolyBlend
============
*/
void R_PolyBlend (void)
{
	if (!gl_polyblend.value)
		return;
	if (!v_blend[3])
		return;

	// this happens in the 3d view so we need to set up blending
	glEnable (GL_BLEND);
	glDisable (GL_DEPTH_TEST);
	glDisable (GL_TEXTURE_2D);

    glLoadIdentity ();

    glRotatef (-90,  1, 0, 0);	    // put Z going up
    glRotatef (90,  0, 0, 1);	    // put Z going up

	glColor4fv (v_blend);

	glBegin (GL_QUADS);

	glVertex3f (10, 100, 100);
	glVertex3f (10, -100, 100);
	glVertex3f (10, -100, -100);
	glVertex3f (10, 100, -100);
	glEnd ();

	glEnable (GL_TEXTURE_2D);
	glDisable (GL_BLEND);
}


int SignbitsForPlane (mplane_t *out)
{
	int	bits, j;

	// for fast box on planeside test

	bits = 0;
	for (j=0 ; j<3 ; j++)
	{
		if (out->normal[j] < 0)
			bits |= 1<<j;
	}
	return bits;
}


void R_SetFrustum (void)
{
	int		i;

	// rotate VPN right by FOV_X/2 degrees
	RotatePointAroundVector (frustum[0].normal, vup, vpn, -(90-r_refdef.fov_x / 2));

	// rotate VPN left by FOV_X/2 degrees
	RotatePointAroundVector (frustum[1].normal, vup, vpn, 90-r_refdef.fov_x / 2);

	// rotate VPN up by FOV_X/2 degrees
	RotatePointAroundVector (frustum[2].normal, vright, vpn, 90-r_refdef.fov_y / 2);

	// rotate VPN down by FOV_X/2 degrees
	RotatePointAroundVector (frustum[3].normal, vright, vpn, -( 90 - r_refdef.fov_y / 2));

	for (i = 0; i < 4; i++)
	{
		frustum[i].type = PLANE_ANYZ;
		frustum[i].dist = DotProduct (r_origin, frustum[i].normal);
		frustum[i].signbits = SignbitsForPlane (&frustum[i]);
	}
}



/*
===============
R_SetupFrame
===============
*/
void R_SetupFrame (void)
{
	int				edgecount;
	vrect_t			vrect;
	float			w, h;

	R_AnimateLight ();

	r_framecount++;

	// build the transformation matrix for the given view angles
	VectorCopy (r_refdef.vieworg, r_origin);

	AngleVectors (r_refdef.viewangles, vpn, vright, vup);

	// current viewleaf
	r_oldviewleaf = r_viewleaf;
	r_viewleaf = Mod_PointInLeaf (r_origin, cl.worldmodel);

	V_SetContentsColor (r_viewleaf->contents);
	V_CalcBlend ();

	c_brush_polys = 0;
	c_alias_polys = 0;

}


void R_SetGLViewport (int x, int y, int w, int h, qboolean clear)
{
	float	screenaspect;
	float	yfov;
	int		i;
	extern	int glwidth, glheight;

	// set up viewpoint
	glViewport (x, y, w, h);

	glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();
	gluPerspective (r_refdef.fov_y,  (float) glwidth / (float) glheight,  4,  4096);

	glCullFace(GL_FRONT);

	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity ();

    glRotatef (-90,  1, 0, 0);	    // put Z going up
    glRotatef (90,  0, 0, 1);	    // put Z going up

    glRotatef (-r_refdef.viewangles[2],  1, 0, 0);
    glRotatef (-r_refdef.viewangles[0],  0, 1, 0);
    glRotatef (-r_refdef.viewangles[1],  0, 0, 1);
    glTranslatef (-r_refdef.vieworg[0],  -r_refdef.vieworg[1],  -r_refdef.vieworg[2]);

	if (gl_cull.value)
		glEnable (GL_CULL_FACE);
	else
		glDisable (GL_CULL_FACE);

	glDisable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);
	glEnable(GL_DEPTH_TEST);

	if (clear)
	{
		// clear down the viewport
		// the stencil buffer is "free" if we have a depth buffer, so no harm in clearing that too...
		glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

		// set up depth function
		glDepthFunc (GL_LEQUAL);
		glDepthRange (0.3, 1);
	}
}


/*
================
R_RenderView

r_refdef must be set before the first call
================
*/
double old_realtime = 0;

void R_RenderView (void)
{
	double	time1, time2;

	if (r_norefresh.value)
		return;

	if (!r_worldentity.model || !cl.worldmodel)
		Sys_Error ("R_RenderView: NULL worldmodel");

	// set client frametime
	// base this on realtime cos cl.time doesn't advance when paused, which can cause some weirdness when effects
	// based on cl_frametime update after being unpaused...
	cl_frametime = realtime - old_realtime;
	old_realtime = realtime;

	if (r_speeds.value)
	{
		glFinish ();
		time1 = Sys_FloatTime ();
		c_brush_polys = 0;
		c_alias_polys = 0;
	}
	else glFinish ();

	// draw this on frame 3 cos the first few are skipped over
	if (r_framecount == 3)
	{
		// i like this...
		SCR_CenterPrint (cl.levelname);
	}

	R_TextureFrame ();

	R_SetupFrame ();

	R_SetFrustum ();

	// done here so we know if we're in water
	R_MarkLeaves ();

	// adds static entities to the list
	R_DrawWorld ();

	// don't let sound get messed up if going slow
	S_ExtraUpdate ();

	// changed the drawing order from here onwards to ensure that translucency doesn't mess things up
	R_DrawEntitiesOnList ();

	R_DrawWaterSurfaces ();

	R_DrawParticles ();

	R_RenderSky ();

	R_DrawViewModel ();

	R_PolyBlend ();

	if (r_speeds.value)
	{
		time2 = Sys_FloatTime ();
		Con_Printf ("%3i ms  %4i wpoly %4i epoly\n", (int)((time2-time1)*1000), c_brush_polys, c_alias_polys); 
	}
}
