/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_misc.c

#include "quakedef.h"

// needed for texture bitmap resource ids
#include "resource.h"


void R_MakeColouredTexture (gltexture_t *glt, int r, int g, int b)
{
	byte data[8][8][4];
	int i, j;

	glt->texnum = texture_extension_number;
	texture_extension_number++;

	for (i = 0; i < 8; i++)
	{
		for (j = 0; j < 8; j++)
		{
			data[i][j][0] = r;
			data[i][j][1] = g;
			data[i][j][2] = b;
			data[i][j][3] = 255;
		}
	}

	glBindTexture (GL_TEXTURE_2D, glt->texnum);

	// mipmap these
	glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);

	// upload the correct texture data without any lightscaling interference
	glTexImage2D (GL_TEXTURE_2D, 
		0, 
		GL_RGBA, 
		8, 
		8, 
		0, 
		GL_RGBA, 
		GL_UNSIGNED_BYTE, 
		data);

	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY, GL_Caps.GL_MaxAnisotropy / 2.0f);
}


void R_InitBloomTexture (void)
{
	byte *data;

	data = malloc (BLOOM_SIZE * BLOOM_SIZE * 4);

	r_bloomtexture.texnum = texture_extension_number;
	texture_extension_number++;

	glBindTexture (GL_TEXTURE_2D, r_bloomtexture.texnum);

	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	// mipmap these
	glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_FALSE);

	// upload the correct texture data without any lightscaling interference
	glTexImage2D (GL_TEXTURE_2D, 
		0, 
		GL_RGBA, 
		BLOOM_SIZE, 
		BLOOM_SIZE, 
		0, 
		GL_RGBA, 
		GL_UNSIGNED_BYTE, 
		data);

	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	free (data);
}


/*
==================
R_InitTextures
==================
*/
void R_InitSkyTextures (void);

void R_InitTextures (void)
{
	// this does nothing!!!
}

// no texture
texture_t r_notexture_mip;
gltexture_t notexture;
gltexture_t r_crosshairtexture;
gltexture_t r_particletexture;
gltexture_t r_blacktexture;
gltexture_t r_whitetexture;
gltexture_t r_bloomtexture;

// create a circular bloom table to prevent squaring of the blooms
byte r_bloomtable[BLOOM_SIZE][BLOOM_SIZE];

void R_InitBloomTable (void)
{
	int x, y;
	int LRADIUS = BLOOM_SIZE / 2;

	for (y = 0; y < BLOOM_SIZE; y++)
	{
		for (x = 0; x < BLOOM_SIZE; x++)
		{
			// find x and y as a -LRADIUS to LRADIUS scale
			int TheRX = x - LRADIUS;
			int TheRY = y - LRADIUS;
			float TheXResult;
			float TheYResult;
			float TheFinalResult;
			unsigned char OutResult;

			// divide X by LRADIUS and square it
			TheXResult = ((float) TheRX / (float) LRADIUS);
			TheXResult *= TheXResult;

			// divide Y by LRADIUS and square it
			TheYResult = ((float) TheRY / (float) LRADIUS);
			TheYResult *= TheYResult;

			// add the 2
			TheFinalResult = TheXResult + TheYResult;

			// clamp 0 to 1
			if (TheFinalResult < 0) TheFinalResult = 0;
			if (TheFinalResult > 1) TheFinalResult = 1;

			// store
			r_bloomtable[x][y] = (byte) ((1.0f - TheFinalResult) * 255.0f);
		}
	}
}


void R_InitResourceTextures (void)
{
	// no texture
	// need to mipmap this one cos it could be used on the world
	r_notexture_mip.alternate_anims = NULL;
	r_notexture_mip.anim_total = 0;
	r_notexture_mip.glt = &notexture;
	strcpy (r_notexture_mip.name, "***notexture***");
	GL_LoadBitmapResourceTexture (&notexture, IDB_BITMAP2, (TEX_MIPMAP));

	// now load the rest of them
	GL_LoadBitmapResourceTexture (&r_crosshairtexture, IDB_BITMAP3, (TEX_MIPMAP | TEX_ALPHAMASK));
	GL_LoadBitmapResourceTexture (&r_particletexture, IDB_BITMAP4, (TEX_MIPMAP | TEX_ALPHAMASK));

	// make various coloured textures
	R_MakeColouredTexture (&r_blacktexture, 0, 0, 0);
	R_MakeColouredTexture (&r_whitetexture, 255, 255, 255);

	// bloom dummy texture
	R_InitBloomTexture ();

	// bloom attenuation table
	R_InitBloomTable ();
}


qboolean R_SetSkyBox (char *name);

void R_Loadsky_f (void)
{
	if (Cmd_Argc () != 2)
	{
		Con_Printf ("loadsky <skyname> : loads the specified skybox\n");
		return;
	}

	if (!R_SetSkyBox (Cmd_Argv (1)))
	{
		Con_Printf ("Failed to load %s\n", Cmd_Argv (1));
		return;
	}

	Con_Printf ("Loaded %s OK\n", Cmd_Argv (1));
}

/*
===============
R_Init
===============
*/
void R_Init (void)
{	
	extern byte *hunk_base;

	Cmd_AddCommand ("timerefresh", R_TimeRefresh_f);	
	Cmd_AddCommand ("pointfile", R_ReadPointFile_f);	
	Cmd_AddCommand ("loadsky", R_Loadsky_f);	

	Cvar_RegisterVariable (&r_norefresh);
	Cvar_RegisterVariable (&r_drawentities);
	Cvar_RegisterVariable (&r_drawviewmodel);
	Cvar_RegisterVariable (&r_novis);
	Cvar_RegisterVariable (&r_speeds);

	Cvar_RegisterVariable (&gl_cull);
	Cvar_RegisterVariable (&gl_polyblend);
	Cvar_RegisterVariable (&gl_nocolors);

	Cvar_RegisterVariable (&gl_doubleeyes);

	R_InitParticles ();

	playertextures = texture_extension_number;
	texture_extension_number += 16;

	// resource and other hard-coded texes
	R_InitResourceTextures ();

	// sky
	R_InitSkyTextures ();
}


/*
===============
R_TranslatePlayerSkin

Translates a skin texture by the per-player color lookup
===============
*/
void R_TranslatePlayerSkin (int playernum)
{
	int		top, bottom;
	byte	translate[256];
	unsigned	translate32[256];
	int		i, j, s;
	model_t	*model;
	aliashdr_t *paliashdr;
	byte	*original;
	unsigned	pixels[512*256], *out;
	unsigned	scaled_width, scaled_height;
	int			inwidth, inheight;
	byte		*inrow;
	unsigned	frac, fracstep;
	extern	byte		**player_8bit_texels_tbl;
	byte	translated[320*200];
	gltexture_t *glt;

	top = cl.scores[playernum].colors & 0xf0;
	bottom = (cl.scores[playernum].colors &15)<<4;

	for (i=0 ; i<256 ; i++)
		translate[i] = i;

	for (i=0 ; i<16 ; i++)
	{
		if (top < 128)	// the artists made some backwards ranges.  sigh.
			translate[TOP_RANGE+i] = top+i;
		else
			translate[TOP_RANGE+i] = top+15-i;
				
		if (bottom < 128)
			translate[BOTTOM_RANGE+i] = bottom+i;
		else
			translate[BOTTOM_RANGE+i] = bottom+15-i;
	}

	// locate the original skin pixels
	currententity = &cl_entities[1+playernum];
	model = currententity->model;
	if (!model)
		return;		// player doesn't have a model yet
	if (model->type != mod_alias)
		return; // only translate skins on alias models

	paliashdr = (aliashdr_t *)Mod_Extradata (model);
	s = paliashdr->skinwidth * paliashdr->skinheight;
	if (currententity->skinnum < 0 || currententity->skinnum >= paliashdr->numskins) {
		Con_Printf("(%d): Invalid player skin #%d\n", playernum, currententity->skinnum);
		original = (byte *)paliashdr + paliashdr->texels[0];
	} else
		original = (byte *)paliashdr + paliashdr->texels[currententity->skinnum];
	if (s & 3)
		Sys_Error ("R_TranslateSkin: s&3");

	inwidth = paliashdr->skinwidth;
	inheight = paliashdr->skinheight;

	for (i = 0; i < s; i += 4)
	{
		translated[i] = translate[original[i]];
		translated[i+1] = translate[original[i+1]];
		translated[i+2] = translate[original[i+2]];
		translated[i+3] = translate[original[i+3]];
	}

	glt = GL_FindGLT (playertextures + playernum);

	if (glt) GL_UploadTexture (glt, translated, paliashdr->skinwidth, paliashdr->skinheight, 8, (TEX_MIPMAP | TEX_ALIAS | TEX_LUMASPEC));
}


// we have no easy way of getting at the worldtype from client side, so let's do a ghastly vile hack
int worldtype;

char *Mod_FindWorldSpawnEntry (model_t *mod, char *KeyToFind)
{
	char *entlump = mod->entities;
	char WSQuotedKey[128];
	static char WSValue[128];
	int seekOffset;
	int i;

	// build the key we wanna find and get the seekoffset
	// we add the quotes around it to ensure that we get an exact match, i.e. so that light for example
	// (not that you'd have light in worldspawn, but you know what i mean) isn't reported as light_blah
	sprintf (WSQuotedKey, "\"%s\"", KeyToFind);
	seekOffset = strlen (WSQuotedKey) + 1;

	// clear to NULLs
	memset (WSValue, 0, sizeof (WSValue));

	while (1)
	{
		// end of the worldspawn entity
		if (*entlump == '}') return WSValue;

		if (!strncmp (entlump, WSQuotedKey, seekOffset - 1))
			goto Eat_Your_Heart_Out_Bono;

		entlump++;

		continue;

Eat_Your_Heart_Out_Bono:;
		// I've just found what I'm looking for!!!!
		// advance by the seekOffset
		entlump += seekOffset;

		// skip whitespace
		while (1)
		{
			if (*entlump == '\"') break;
			entlump++;
		}

		// skip the quote we just found
		entlump++;

		// copy to WSValue
		for (i = 0; i < 127; i++)
		{
			if (*entlump == '\"') break;

			WSValue[i] = *entlump;
			entlump++;
		}

		return WSValue;
	}

	// never gets here, just keep the compiler happy
	return WSValue;
}


/*
===============
R_NewMap
===============
*/
void GL_BuildLightmaps (void);
void GL_BuildSurfaces (void);

void R_NewMap (void)
{
	int		i;

	// clear light styles
	for (i = 0; i < 256; i++)
		d_lightstylevalue[i] = 264;

	// clear world entity and set it's model to the new map
	memset (&r_worldentity, 0, sizeof (r_worldentity));
	r_worldentity.model = cl.worldmodel;

	// clear out efrags in case the level hasn't been reloaded
	// FIXME: is this one short?  FIXED - yes.
	for (i = 0; i < cl.worldmodel->numleafs + 1; i++)
		cl.worldmodel->leafs[i].efrags = NULL;

	// no viewleaf
	r_viewleaf = NULL;

	// clear down particles
	R_ClearParticles ();

	// set up memory for alias model rendering
	AllocAliasPolyVerts ();

	// make lightmaps and polygons
	GL_BuildLightmaps ();

	// the lightmaps must be built before this in order to get light s and t for the texcoords
	GL_BuildSurfaces ();

	// identify sky texture
	skytexturenum = -1;

	// set up texture chains here also
	for (i = 0; i < cl.worldmodel->numtextures; i++)
	{
		if (!cl.worldmodel->textures[i]) continue;

		if (!Q_strncmp (cl.worldmodel->textures[i]->name,"sky", 3)) skytexturenum = i;

		cl.worldmodel->textures[i]->texturechain = NULL;
	}

	// find the worldtype
	// hack - fix for wrong worldtype in dm2
	if (!strcmp (cl.levelname, "Claustrophobopolis"))
		worldtype = 1;
	else
	{
		// hmmm - what happens if we send a NULL into atoi here...
		// FIXED - modded Mod_FindWorldSpawnEntry so it never returns NULL
		worldtype = atoi (Mod_FindWorldSpawnEntry (cl.worldmodel, "worldtype"));
	}

	// gib tracking
	numgibsonlist = 0;
}


/*
====================
R_TimeRefresh_f

For program optimization
====================
*/
void R_TimeRefresh_f (void)
{
	int			i;
	float		start, stop, time;
	int			startangle;
	vrect_t		vr;

	glDrawBuffer  (GL_FRONT);
	glFinish ();

	start = Sys_FloatTime ();
	for (i=0 ; i<128 ; i++)
	{
		r_refdef.viewangles[1] = i/128.0*360.0;
		R_RenderView ();
	}

	glFinish ();
	stop = Sys_FloatTime ();
	time = stop-start;
	Con_Printf ("%f seconds (%f fps)\n", time, 128/time);

	glDrawBuffer  (GL_BACK);
	GL_EndRendering ();
}

