/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_surf.c: surface-related refresh code

#include "quakedef.h"

// we're only allowed 1024 textures in glquake so this is safe
texture_t *texturelist[1024];

int			skytexturenum;

void R_PushDlights (mnode_t *headnode);

#define SINTABLE_SIZE 1024
float sintable[SINTABLE_SIZE];
int STBitMask;
float STMult;


// gl_rlight.c imports required by code move
void R_ModifyLightmaps (void);
void R_CheckLightmapModification (msurface_t *surf);
void R_InitLight (void);

// liquid surface subdivision
int GL_SubdivideSurface (model_t *mod, msurface_t *fa, int waterverts);

// vertexes for drawing
polyverts_t *polyverts = NULL;


// frame render state
typedef struct framerenderstate_s
{
	// frame has translucent liquid
	qboolean alphaliquid;

	// frame has non-translucent liquid
	// pedants run for cover!!!
	qboolean solidliquid;

	// frame has sky
	qboolean sky;

	// frame has lumas
	qboolean luma;
} framerenderstate_t;

framerenderstate_t framerenderstate;


// blooms
cvar_t r_bloom = {"r_bloom", "0", true};
cvar_t r_bloomalpha = {"r_bloomalpha", "0.025", true};
cvar_t r_bloomradius = {"r_bloomradius", "6", true};
extern byte r_bloomtable[BLOOM_SIZE][BLOOM_SIZE];

/*
===============
R_TextureAnimation

Returns the proper texture for a given time and base texture
===============
*/
texture_t *R_TextureAnimation (texture_t *base)
{
	int		reletive;
	int		count;

	// store the original base to avoid errors
	texture_t *original = base;

	if (currententity->frame)
	{
		if (base->alternate_anims)
			base = base->alternate_anims;
	}
	
	if (!base->anim_total)
		return base;

	reletive = (int)(cl.time*10) % base->anim_total;

	count = 0;	

	while (base->anim_min > reletive || base->anim_max <= reletive)
	{
		base = base->anim_next;

		// return the original tex on an error condition
		// (broken cycle/infinite cycle)
		if (!base) return original;
		if (++count > 100) return original;
	}

	return base;
}


/*
================
R_GenTurbTile
================
*/
#if 0
void R_GenTurbTile (pixel_t *pbasetex, void *pdest)
{
	int		*turb;
	int		i, j, s, t;
	byte	*pd;
	
	turb = sintable + ((int)(cl.time*SPEED)&(CYCLE-1));
	pd = (byte *)pdest;

	for (i=0 ; i<TILE_SIZE ; i++)
	{
		for (j=0 ; j<TILE_SIZE ; j++)
		{	
			s = (((j << 16) + turb[i & (CYCLE-1)]) >> 16) & 63;
			t = (((i << 16) + turb[j & (CYCLE-1)]) >> 16) & 63;
			*pd++ = *(pbasetex + (t<<6) + s);
		}
	}
}
#else
void R_GenTurbTile (byte *basetex, byte *dest)
{
}
#endif


/*
=============================================================

	BRUSH MODELS

=============================================================
*/


/*
================
R_DrawWaterSurfaces
================
*/
void R_DrawWaterSurfaces (void)
{
	int			i;
	msurface_t	*s;
	texture_t	*t;

	// go back to the world matrix - do this irrespective of whether or not we have liquid
	glLoadMatrixf (cl_entities[0].matrix);

	// no liquid to draw
	if (!framerenderstate.alphaliquid && !framerenderstate.solidliquid)
		return;

	// let's do the timewarp again
	GL_EnableVertexArrays (polyverts, VA_ST);

	// scale these on the texture matrix
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glScalef (0.015625, 0.015625, 1);

	// draw non-translucent surfaces first
	for (i = 0; i < 1024 && framerenderstate.solidliquid; i++)
	{
		// end of the line
		if (!texturelist[i]) break;

		// not liquid
		if (!(texturelist[i]->glt->flags & TEX_LIQUID)) continue;

		// get the chained surfs
		s = texturelist[i]->texturechain;

		// no chained surfs!!!
		if (!s) continue;

		// to be sure, to be sure
		if (!(s->flags & SURF_DRAWTURB)) continue;

		// non-translucent only
		if ((s->flags & SURF_DRAWLAVA) || (s->flags & SURF_DRAWTELE) || !cl.worldmodel->transwater)
		{
			// bind texture
			glBindTexture (GL_TEXTURE_2D, texturelist[i]->glt->texnum);

			// draw the chain in the normal fashion
			for (; s; s = s->texturechain)
			{
				// draw the surf
				GL_DrawVertexArrays (s->polys);
			}
		}
	}

	if (framerenderstate.alphaliquid)
	{
		glEnable (GL_BLEND);
		glDepthMask (GL_FALSE);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

		for (i = 0; i < 1024; i++)
		{
			// end of the line
			if (!texturelist[i]) break;

			// not liquid
			if (!(texturelist[i]->glt->flags & TEX_LIQUID)) continue;

			// get the chained surfs
			s = texturelist[i]->texturechain;

			// no chained surfs!!!
			if (!s) continue;

			// to be sure, to be sure
			if (!(s->flags & SURF_DRAWTURB)) continue;

			// non-alpha
			if (s->flags & SURF_DRAWLAVA) continue;
			if (s->flags & SURF_DRAWTELE) continue;

			// bind texture
			glBindTexture (GL_TEXTURE_2D, texturelist[i]->glt->texnum);

			// set translucent colour according to the liquid type
			if (s->flags & SURF_DRAWSLIME)
				glColor4f (1, 1, 1, 0.75);
			else glColor4f (1, 1, 1, 0.5);

			// draw the chain in the normal fashion
			for (; s; s = s->texturechain)
			{
				// draw the surf
				GL_DrawVertexArrays (s->polys);
			}
		}

		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glColor4f (1, 1, 1, 1);
		glDepthMask (GL_TRUE);
		glDisable (GL_BLEND);
	}

	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);

	GL_DisableVertexArrays ();
}


/*
================
R_AddTextureToList

================
*/
void R_AddTextureToList (texture_t *tx)
{
	int i;

	// always keep 1 free slot
	for (i = 0; i < 1023; i++)
	{
		// add to the first free slot
		if (!texturelist[i])
		{
			texturelist[i] = tx;
			return;
		}

		// already added
		if (texturelist[i]->glt->texnum == tx->glt->texnum) return;
	}

	Sys_Error ("Texture list overflow!");
}


/*
================
R_ClearTextureChains

================
*/
void R_ClearTextureChains (void)
{
	int i;

	for (i = 0; i < 1024; i++)
	{
		if (!texturelist[i]) break;

		texturelist[i]->texturechain = NULL;
	}
}


/*
================
R_DrawTextureChains

================
*/
void R_DrawTextureChains (void)
{
	int i;
	msurface_t *s;
	qboolean lumaPickup = false;
	int old_lightmap = -1;

	// bring up multitexturing
	glActiveTexture (GL_TEXTURE1);
	glEnable (GL_TEXTURE_2D);

	// combine for proper lighting level
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
	glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PREVIOUS);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_TEXTURE);
	glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 2.0f);

	// use a 3 TMU render for the fullbrights if available
	if (GL_Caps.GL_TextureUnits > 2)
	{
		// bring up TMU2 and set add environment
		glActiveTexture (GL_TEXTURE2);
		glEnable (GL_TEXTURE_2D);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_ADD);

		// bring up vertex arrays for 3 TMUs
		GL_EnableVertexArrays (polyverts, VA_STLMFB);
	}
	else
	{
		// bring up vertex arrays for 2 TMUs
		GL_EnableVertexArrays (polyverts, VA_STLM);
	}

	for (i = 0; i < 1024; i++)
	{
		// end of the line
		if (!texturelist[i]) break;

		// get the chained surfs
		s = texturelist[i]->texturechain;

		// no chained surfs!!!
		if (!s) continue;

		// deal with these later
		if (s->flags & SURF_DRAWTURB) continue;
		if (s->flags & SURF_DRAWSKY) continue;

		// activate tmu 0 for the world and bind the texture
		glActiveTexture (GL_TEXTURE0);
		glBindTexture (GL_TEXTURE_2D, texturelist[i]->glt->texnum);

		if (GL_Caps.GL_TextureUnits > 2)
		{
			// bind luma in TMU 2
			// lumapickup will never be true
			glActiveTexture (GL_TEXTURE2);

			// this is set to the black texnum if there is no luma to keep the blend good.
			glBindTexture (GL_TEXTURE_2D, texturelist[i]->glt->lumatexnum);
		}
		else
		{
			// check for luma pickup
			if (texturelist[i]->glt->lumatexnum) lumaPickup = true;
		}

		// activate tmu 1 for the lightmaps
		glActiveTexture (GL_TEXTURE1);

		// draw the chain
		for (; s; s = s->texturechain)
		{
			// only bind the lightmap texture if it changes
			if (s->gllmtexnum != old_lightmap)
			{
				glBindTexture (GL_TEXTURE_2D, s->gllmtexnum);
				old_lightmap = s->gllmtexnum;
			}

			// draw the surf
			GL_DrawVertexArrays (s->polys);
		}
	}

	// take down multitexturing
	// tmu 1 will be always active coming out to here so take it straight down
	glDisable (GL_TEXTURE_2D);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	// restore colour scale in TMU1
	glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 1.0f);

	// shut down TMU2 if required
	if (GL_Caps.GL_TextureUnits > 2)
	{
		glActiveTexture (GL_TEXTURE2);
		glDisable (GL_TEXTURE_2D);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	}

	// revert to TMU0
	glActiveTexture (GL_TEXTURE0);

	// bring down vertex arrays
	GL_DisableVertexArrays ();

	// no pickups to do
	if (!lumaPickup) return;

	// enable blending for pickup passes
	glEnable (GL_BLEND);
	glDepthMask (GL_FALSE);

	if (lumaPickup)
	{
		// add in fullbrights
		GL_EnableVertexArrays (polyverts, VA_ST);

		glBlendFunc (GL_ONE, GL_ONE);

		for (i = 0; i < 1024; i++)
		{
			// end of the line
			if (!texturelist[i]) break;

			// no fullbright
			if (!texturelist[i]->glt->lumatexnum) continue;

			// get the chained surfs
			s = texturelist[i]->texturechain;

			// no chained surfs!!!
			if (!s) continue;

			// no luma on these!!!
			if (s->flags & SURF_DRAWTURB) continue;
			if (s->flags & SURF_DRAWSKY) continue;

			glBindTexture (GL_TEXTURE_2D, texturelist[i]->glt->lumatexnum);

			// draw the chain
			for (; s; s = s->texturechain)
			{
				// draw the surf
				GL_DrawVertexArrays (s->polys);
			}
		}

		GL_DisableVertexArrays ();
	}

	// back to regular, unblended drawing
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDepthMask (GL_TRUE);
	glDisable (GL_BLEND);
}


void R_WarpSurface (msurface_t *surf)
{
	int i;
	glpoly_t *poly = surf->polys;
	float st[2];

	for (; poly; poly = poly->next)
	{
		for (i = 0; i < poly->numverts; i++)
		{
			// derive the texcoords from the verts!!!
			st[0] = poly->verts[i].tv[surf->warpindex[0]];
			st[1] = poly->verts[i].tv[surf->warpindex[1]];

			// you're warped, you are!
			// do the warp in the 64 unit texcoords
			poly->verts[i].st[0] = st[0] + sintable[(int) ((st[1] * 0.025 + cl.time) * STMult) & STBitMask];
			poly->verts[i].st[1] = st[1] + sintable[(int) ((st[0] * 0.025 + cl.time) * STMult) & STBitMask];
		}
	}
}


/*
=================
R_AddSurfaceToChain

Adds a surface to the correct texture chain.
=================
*/
void R_AddSurfaceToChain (msurface_t *surf)
{
	texture_t *tx;

	// check render state
	if (surf->flags & SURF_DRAWLAVA)
		framerenderstate.solidliquid = true;
	else if (surf->flags & SURF_DRAWTELE)
		framerenderstate.solidliquid = true;
	else if (surf->flags & SURF_DRAWTURB && cl.worldmodel->transwater)
		framerenderstate.alphaliquid = true;
	else if (surf->flags & SURF_DRAWTURB)
		framerenderstate.solidliquid = true;
	else if (surf->flags & SURF_DRAWSKY)
		framerenderstate.sky = true;

	// check for lightmap modification
	R_CheckLightmapModification (surf);

	// get the correct animation
	tx = R_TextureAnimation (surf->texinfo->texture);

	// add to the chain
	surf->texturechain = tx->texturechain;
	tx->texturechain = surf;

	// the following checks require the tx to have a glt
	if (!(surf->flags & SURF_DRAWSKY))
	{
		// mark if we have lumas
		if (tx->glt->lumatexnum) framerenderstate.luma = true;
	}

	if (surf->flags & SURF_DRAWTURB) R_WarpSurface (surf);
}


/*
=================
R_AddTransformedBModelToList

Adds an inline ("*", contained in the world BSP) model to the texture chains for drawing
=================
*/
void R_AddTransformedBModelToList (entity_t *e)
{
	int			j, k;
	vec3_t		mins, maxs;
	int			i, numsurfaces;
	msurface_t	*surf;
	float		dot;
	mplane_t	*pplane;
	model_t		*clmodel;
	qboolean	rotated;
	vec3_t		originhack;
	float		matrix[16];

	currententity = e;

	clmodel = e->model;

	if (e->angles[0] || e->angles[1] || e->angles[2])
	{
		rotated = true;
		for (i = 0; i < 3; i++)
		{
			mins[i] = e->origin[i] - clmodel->radius;
			maxs[i] = e->origin[i] + clmodel->radius;
		}
	}
	else
	{
		rotated = false;
		VectorAdd (e->origin, clmodel->mins, mins);
		VectorAdd (e->origin, clmodel->maxs, maxs);
	}

	// outside the view frustum
	if (R_CullBox (mins, maxs)) return;

	VectorSubtract (r_refdef.vieworg, e->origin, modelorg);

	if (rotated)
	{
		vec3_t	temp;
		vec3_t	forward, right, up;

		VectorCopy (modelorg, temp);
		AngleVectors (e->angles, forward, right, up);
		modelorg[0] = DotProduct (temp, forward);
		modelorg[1] = -DotProduct (temp, right);
		modelorg[2] = DotProduct (temp, up);
	}

	surf = &clmodel->surfaces[clmodel->firstmodelsurface];

	// calculate dynamic lighting for bmodel if it's not an
	// instanced model
	R_PushDlights (clmodel->nodes + clmodel->hulls[0].firstclipnode);

	// hack the origin to prevent bmodel z fighting
	VectorCopy (e->origin, originhack);
	e->origin[0] -= DIST_EPSILON;
	e->origin[1] -= DIST_EPSILON;
	e->origin[2] -= DIST_EPSILON;

	// angle hack ("stupid quake bug")
	e->angles[0] = -e->angles[0];

	// transform
	R_RotateForEntity (e);

	// add to the list for drawing
	for (i = 0; i < clmodel->nummodelsurfaces; i++, surf++)
	{
		// find which side of the node we are on
		pplane = surf->plane;

		dot = DotProduct (modelorg, pplane->normal) - pplane->dist;

		if (((surf->flags & SURF_PLANEBACK) && (dot < -BACKFACE_EPSILON)) || (!(surf->flags & SURF_PLANEBACK) && (dot > BACKFACE_EPSILON)))
		{
			// transform
			R_TransformPolys (surf->polys, e->matrix);

			// add to the drawing chains
			R_AddSurfaceToChain (surf);
		}
	}

	// un-hack the origin
	VectorCopy (originhack, e->origin);

	// un-hack the angles ("stupid quake bug")
	e->angles[0] = -e->angles[0];
}


/*
=================
R_AddBModelsToDrawLists

Adds all inline Bmodels to the texture chains for drawing
=================
*/
void R_AddBModelsToDrawLists (void)
{
	int		i;
	entity_t *ent;

	// don't draw any entities
	if (!r_drawentities.value) return;

	// go through all of the currently visible entities
	for (i = 0; i < cl_numvisedicts; i++)
	{
		// get the current entity
		ent = cl_visedicts[i];

		// entity has no model!!!
		if (!ent->model) continue;

		// check the type
		if (ent->model->type == mod_brush)
		{
			// and the type of bmodel
			if (ent->model->name[0] == '*')
			{
				// inline bmodels are never shared by entities so it's OK to add them to the list
				R_AddTransformedBModelToList (ent);
			}
		}
	}
}


/*
===================
R_DrawBrushModel

Box models
===================
*/
void R_DrawBrushModel (entity_t *e)
{
	msurface_t *surf;
	texture_t *tx;
	int i;
	vec3_t shadecolour;
	vec3_t originhack;

	// set up origin for lighting
	// move it up a bit cos these have their origin at a bottom corner
	originhack[0] = e->origin[0];
	originhack[1] = e->origin[1];
	originhack[2] = e->origin[2] + 8;

	// get lighting
	R_LightPoint (originhack, shadecolour);

	// 0 to 1 scale
	shadecolour[0] /= 128.0f;
	shadecolour[1] /= 128.0f;
	shadecolour[2] /= 128.0f;

	// needed for texture animation
	currententity = e;

	// don't bother lightmapping these
	glColor3fv (shadecolour);

	// draw it all, no need to be sexy here
	for (surf = &e->model->surfaces[e->model->firstmodelsurface], i = 0; i < e->model->nummodelsurfaces; i++, surf++)
	{
		// animation
		tx = R_TextureAnimation (surf->texinfo->texture);

		// bind
		glActiveTexture (GL_TEXTURE0);
		glBindTexture (GL_TEXTURE_2D, tx->glt->texnum);

		glActiveTexture (GL_TEXTURE1);

		if (tx->glt->lumatexnum)
			glBindTexture (GL_TEXTURE_2D, tx->glt->lumatexnum);
		else glBindTexture (GL_TEXTURE_2D, r_blacktexture.texnum);

		// these don't rotate
		R_TranslatePolys (surf->polys, e->origin);

		// draw the surf
		GL_DrawVertexArrays (surf->polys);
	}
}


/*
=============================================================

	WORLD MODEL

=============================================================
*/

/*
================
R_RecursiveWorldNode
================
*/
void R_RecursiveWorldNode (mnode_t *node)
{
	int			i, c, side;
	mplane_t	*plane;
	msurface_t	*surf, **mark;
	mleaf_t		*leaf;
	double		dot;
	texture_t	*tx;

	// solid
	if (node->contents == CONTENTS_SOLID) return;

	// not visible
	if (node->visframe != r_visframecount) return;

	// outside the view frustum
	if (R_CullBox (node->minmaxs, node->minmaxs + 3)) return;

	// if a leaf node, draw stuff
	if (node->contents < 0)
	{
		leaf = (mleaf_t *) node;

		mark = leaf->firstmarksurface;
		c = leaf->nummarksurfaces;

		if (c)
		{
			do
			{
				(*mark)->visframe = r_framecount;
				mark++;
			} while (--c);
		}

		// deal with model fragments in this leaf
		if (leaf->efrags) R_StoreEfrags (&leaf->efrags);

		return;
	}

	// node is just a decision point, so go down the apropriate sides
	// find which side of the node we are on
	plane = node->plane;

	switch (plane->type)
	{
	case PLANE_X:
		dot = modelorg[0] - plane->dist;
		break;
	case PLANE_Y:
		dot = modelorg[1] - plane->dist;
		break;
	case PLANE_Z:
		dot = modelorg[2] - plane->dist;
		break;
	default:
		dot = DotProduct (modelorg, plane->normal) - plane->dist;
		break;
	}

	if (dot >= 0)
		side = 0;
	else
		side = 1;

	// recurse down the children, front side first
	R_RecursiveWorldNode (node->children[side]);

	// draw stuff
	c = node->numsurfaces;

	if (c)
	{
		surf = cl.worldmodel->surfaces + node->firstsurface;

		if (dot < 0 -BACKFACE_EPSILON)
			side = SURF_PLANEBACK;
		else if (dot > BACKFACE_EPSILON)
			side = 0;

		for (; c; c--, surf++)
		{
			// not visible
			if (surf->visframe != r_framecount) continue;

			if (((dot < 0) ^ !!(surf->flags & SURF_PLANEBACK))) continue;

			R_AddSurfaceToChain (surf);
		}
	}

	// recurse down the back side
	R_RecursiveWorldNode (node->children[!side]);
}


void R_QuadBloom (int x, int y)
{
	// draw at front of Z order
	glTexCoord2f (0, 0);
	glVertex3f (x, y, 0);

	glTexCoord2f (0, 1);
	glVertex3f (x, BLOOM_SIZE + y, 0);

	glTexCoord2f (1, 1);
	glVertex3f (BLOOM_SIZE + x, BLOOM_SIZE + y, 0);

	glTexCoord2f (1, 0);
	glVertex3f (BLOOM_SIZE + x, y, 0);
}


qboolean R_AttenuateBloom (int x, int y)
{
	byte alpha;

	// rescale 0 to (r_bloomradius.value * 2 + 1) 
	x += r_bloomradius.value;
	y += r_bloomradius.value;

	// rescale to bloomtable dimensions
	x *= BLOOM_SIZE;
	y *= BLOOM_SIZE;

	// get correct offsets in the table
	x /= (r_bloomradius.value * 2) + 1;
	y /= (r_bloomradius.value * 2) + 1;

	// paranoid clamping
	if (x >= BLOOM_SIZE) x = BLOOM_SIZE - 1;
	if (y >= BLOOM_SIZE) y = BLOOM_SIZE - 1;

	// full attenuation = no bloom
	if (r_bloomtable[x][y] == 0)
		return false;

	// modulate the alpha by the lookup
	alpha = (byte) (r_bloomalpha.value * r_bloomtable[x][y]);

	// set colour
	glColor3ub (alpha, alpha, alpha);

	// success
	return true;
}


void R_BloomToTexture (void)
{
	int i;
	int x, y;
	int minbloom, maxbloom;
	msurface_t *s;

	// blooms are disabled
	if (!r_bloom.value) return;

	// nothing to bloom!
	if (!framerenderstate.luma) return;

	// cvar sanity
	if (r_bloomalpha.value < 0) r_bloomalpha.value = 0;
	if (r_bloomalpha.value > 1) r_bloomalpha.value = 1;
	if (r_bloomradius.value < 1) r_bloomradius.value = 1;
	if (r_bloomradius.value > BLOOM_SIZE / 2) r_bloomradius.value = BLOOM_SIZE / 2;

	// bloom dimensions
	minbloom = -r_bloomradius.value;
	maxbloom = r_bloomradius.value + 1;

	// create viewport for bloom rendering
	R_SetGLViewport (0, 0, BLOOM_SIZE, BLOOM_SIZE, true);

	// add in fullbrights
	GL_EnableVertexArrays (polyverts, VA_ST);

	for (i = 0; i < 1024; i++)
	{
		// end of the line
		if (!texturelist[i]) break;

		// get the chained surfs
		s = texturelist[i]->texturechain;

		// no chained surfs!!!
		if (!s) continue;

		// no luma on these!!!
		if (s->flags & SURF_DRAWTURB) continue;
		if (s->flags & SURF_DRAWSKY) continue;

		// bind the black texture if we have no lumas.  this is important, as we won't get proper
		// depthmasking otherwise.
		if (texturelist[i]->glt->lumatexnum)
			glBindTexture (GL_TEXTURE_2D, texturelist[i]->glt->lumatexnum);
		else glBindTexture (GL_TEXTURE_2D, r_blacktexture.texnum);

		// draw the chain
		for (; s; s = s->texturechain)
		{
			// draw the surf
			GL_DrawVertexArrays (s->polys);
		}
	}

	GL_DisableVertexArrays ();

	// copy to bloom texture
	glBindTexture (GL_TEXTURE_2D, r_bloomtexture.texnum);
	glCopyTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, 0, 0, BLOOM_SIZE, BLOOM_SIZE);

	// set up 2D viewport
	glViewport (0, 0, BLOOM_SIZE, BLOOM_SIZE);

	glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();
	glOrtho (0, BLOOM_SIZE, 0, BLOOM_SIZE, -99999, 99999);

	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity ();

	glDisable (GL_DEPTH_TEST);
	glDisable (GL_CULL_FACE);

	// blooms hurt fillrate badly when doing multiple fullscreen overlays, so here we just do the overlay 
	// into the current viewport.  first off, we must clear it to all black.
	glBindTexture (GL_TEXTURE_2D, r_blacktexture.texnum);
	glBegin (GL_QUADS);
	R_QuadBloom (0, 0);
	glEnd ();

	// now set up for the bloom overlay
	// additive blend mode
	glEnable (GL_BLEND);
	glDepthMask (GL_FALSE);
	glBlendFunc (GL_ONE, GL_ONE);

	// bind bloom overlay texture
	glBindTexture (GL_TEXTURE_2D, r_bloomtexture.texnum);

	// modulate with primary colour
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glBegin (GL_QUADS);

	// bloom radius is doubled and we increment by 2 to prevent excessive fillrate hammering
	// fixed - using a 128 * 128 bloom texture is visually acceptable here
	for (x = minbloom; x < maxbloom; x++)
	{
		for (y = minbloom; y < maxbloom; y++)
		{
			// set bloom attenuation
			if (!R_AttenuateBloom (x, y)) continue;

			// draw the quad
			R_QuadBloom (x, y);
		}
	}

	glEnd ();

	// now we read it back again in it's final bloomed format
	glCopyTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, 0, 0, BLOOM_SIZE, BLOOM_SIZE);

	// restore state
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);
	glColor3f (1, 1, 1);
}


void R_BloomOverlay (void)
{
	int x, y;
	int minbloom, maxbloom;

	// blooms are disabled
	if (!r_bloom.value) return;

	// nothing to bloom!
	if (!framerenderstate.luma) return;

	// set up 2D viewport
	glViewport (0, 0, glwidth, glheight);

	glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();
	glOrtho (0, BLOOM_SIZE, 0, BLOOM_SIZE, -99999, 99999);

	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity ();

	glDisable (GL_DEPTH_TEST);
	glDisable (GL_CULL_FACE);

	// additive blend mode
	glEnable (GL_BLEND);
	glDepthMask (GL_FALSE);
	glBlendFunc (GL_ONE, GL_ONE);

	// bind bloom overlay texture
	glBindTexture (GL_TEXTURE_2D, r_bloomtexture.texnum);

	glBegin (GL_QUADS);
	R_QuadBloom (0, 0);
	glEnd ();

	// restore state
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);
	glColor3f (1, 1, 1);

	// restore 3D view (but don't clear the vp)
	R_SetGLViewport (0, 0, glwidth, glheight, false);
}


/*
=============
R_DrawWorld
=============
*/
void R_DrawWorld (void)
{
	entity_t	ent;
	int			i;

	memset (&ent, 0, sizeof(ent));
	ent.model = cl.worldmodel;

	VectorCopy (r_refdef.vieworg, modelorg);

	currententity = &ent;

	// default primary colour
	glColor3f (1, 1, 1);

	// clear down texture chain state and drawlists
	R_ClearTextureChains ();

	// mark dynamic lights
	R_PushDlights (cl.worldmodel->nodes);

	// clear volatile render state
	memset (&framerenderstate, 0, sizeof (framerenderstate_t));

	// add the world to the drawlists
	R_RecursiveWorldNode (cl.worldmodel->nodes);

	// add * models to the drawlists
	R_AddBModelsToDrawLists ();

	// set any lightmap modifications we need
	R_ModifyLightmaps ();

	// build blooms
	R_BloomToTexture ();

	// main viewport for rendering
	R_SetGLViewport (0, 0, glwidth, glheight, true);

	// save out the master modelview matrix
	glGetFloatv (GL_MODELVIEW_MATRIX, cl_entities[0].matrix);

	// draw the surfs
	R_DrawTextureChains ();

	// bloom overlay
	R_BloomOverlay ();
}


/*
===============
R_AddVisibleLeaf

adds PVS for a given leaf to the global PVS
===============
*/
void R_AddVisibleLeaf (mleaf_t *leaf)
{
	byte	*vis;
	mnode_t	*node;
	int		i;
	int		j = 0;

	vis = Mod_LeafPVS (leaf, cl.worldmodel);

	for (i = 0; i < cl.worldmodel->numleafs; i++)
	{
		// moved r_novis check to here
		if ((vis[i >> 3] & (1 << (i & 7))) || r_novis.value)
		{
			j++;
			node = (mnode_t *) &cl.worldmodel->leafs[i + 1];

			do
			{
				if (node->visframe == r_visframecount) break;

				node->visframe = r_visframecount;
				node = node->parent;
			} while (node);
		}
	}

	Con_DPrintf ("%i Visible leafs\n", j);
}


/*
===============
R_MarkLeaves

Mark visible leafs
===============
*/
void R_MarkLeaves (void)
{
	static int OldNoVis = 0;

	// leaf hasn't changed
	if (r_oldviewleaf == r_viewleaf && (int) r_novis.value == OldNoVis) return;

	OldNoVis = (int) r_novis.value;

	// if we don't have an old view leaf, set it to the current so that member checks are valid
	// needed because r_oldviewleaf is NULL on the first frame
	if (!r_oldviewleaf) r_oldviewleaf = r_viewleaf;

	// go to a new pvs
	r_visframecount++;

	// add visibility for the primary viewleaf
	R_AddVisibleLeaf (r_viewleaf);

	// if the contents have changed, add the previous viewleaf to avoid transition effects
	if (r_viewleaf->contents != r_oldviewleaf->contents)
		R_AddVisibleLeaf (r_oldviewleaf);

	// set old = current for checking next frame
	r_oldviewleaf = r_viewleaf;
}



/*
================
GL_BuildPolygonFromSurface

Renamed to same as Q2 cos it's not a display list, is it?
Maybe it was in an early version of the engine, but not no more for sure...
================
*/
int GL_BuildPolygonFromSurface (model_t *mod, msurface_t *surf, int worldverts)
{
	int			i;
	medge_t		*r_pedge;
	float		*vec;
	float		s, t;
	glpoly_t	*poly;

	// allocate space on the hunk for the poly and it's verts
	poly = Hunk_Alloc (sizeof (glpoly_t));
	poly->basev = Hunk_Alloc (sizeof (vec3_t) * surf->numedges);

	poly->next = surf->polys;
	poly->flags = surf->flags;
	surf->polys = poly;
	poly->numverts = surf->numedges;

	poly->vertoffset = worldverts;
	worldverts += poly->numverts;

	for (i = 0; i < poly->numverts; i++)
	{
		if (mod->surfedges[surf->firstedge + i] > 0)
		{
			r_pedge = &mod->edges[mod->surfedges[surf->firstedge + i]];
			vec = mod->vertexes[r_pedge->v[0]].position;
		}
		else
		{
			r_pedge = &mod->edges[-mod->surfedges[surf->firstedge + i]];
			vec = mod->vertexes[r_pedge->v[1]].position;
		}

		// base verts
		poly->basev[i][0] = vec[0];
		poly->basev[i][1] = vec[1];
		poly->basev[i][2] = vec[2];
	}

	// set the most optimum primitive
	switch (poly->numverts)
	{
	case 0:
	case 1:
	case 2:
		Sys_Error ("Poly with less than 3 verts");
		break;

	case 3:
		poly->PRIMITIVE_TYPE = GL_TRIANGLES;
		break;

	case 4:
		poly->PRIMITIVE_TYPE = GL_QUADS;
		break;

	default:
		poly->PRIMITIVE_TYPE = GL_TRIANGLE_FAN;
		break;
	}

	return worldverts;
}


void GL_BuildPolyVerts (msurface_t *surf)
{
	int i;
	float s, t;
	glpoly_t *poly;

	for (poly = surf->polys; poly; poly = poly->next)
	{
		// link in the verts
		poly->verts = &polyverts[poly->vertoffset];

		for (i = 0; i < poly->numverts; i++)
		{
			// verts and transformed verts
			// tv will always be the same as v for the world, but may change for bmodels
			poly->verts[i].tv[0] = poly->basev[i][0];
			poly->verts[i].tv[1] = poly->basev[i][1];
			poly->verts[i].tv[2] = poly->basev[i][2];

			// don't need the rest for water
			if (surf->flags & SURF_DRAWTURB)
				continue;

			// regular texcoords
			s = DotProduct (poly->basev[i], surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3];
			s /= surf->texinfo->texture->width;

			t = DotProduct (poly->basev[i], surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3];
			t /= surf->texinfo->texture->height;

			poly->verts[i].st[0] = s;
			poly->verts[i].st[1] = t;

			// lightmap texcoords
			s = DotProduct (poly->basev[i], surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3];
			s -= surf->texturemins[0];
			s += surf->light_s * 16;
			s += 8;
			s /= LM_BLOCK_WIDTH * 16;

			t = DotProduct (poly->basev[i], surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3];
			t -= surf->texturemins[1];
			t += surf->light_t * 16;
			t += 8;
			t /= LM_BLOCK_HEIGHT * 16;

			poly->verts[i].lm[0] = s;
			poly->verts[i].lm[1] = t;
		}
	}
}


/*
================
GL_BuildSurfaces

Split from GL_BuildLightmaps
================
*/
void GL_BuildSurfaces (void)
{
	int		i, j;
	model_t	*m;
	int worldverts = 0;

	// init the texturelist
	for (i = 0; i < 1024; i++) texturelist[i] = NULL;

	// build polys for all brush models in the cache
	for (j = 1; j < MAX_MODELS; j++)
	{
		m = cl.model_precache[j];
		if (!m) break;

		// somebody in ID forgot this check, meaning that it will attempt to build polys for alias and sprite 
		// models too! (it won't of course, as they have no surfs, but it was still a mistake to forget it)
		if (m->type != mod_brush) continue;

		// add to the texture list for this model
		for (i = 0; i < m->numtextures; i++)
		{
			// NULL texture
			if (!m->textures[i]) continue;

			// sky
			if (!m->textures[i]->glt) continue;

			// add to the list
			R_AddTextureToList (m->textures[i]);

			// if a luma is unavailable, set to the black texture
			if (!m->textures[i]->glt->lumatexnum)
				m->textures[i]->glt->lumatexnum = r_blacktexture.texnum;
		}

		if (m->name[0] == '*') continue;

		for (i = 0; i < m->numsurfaces; i++)
		{
			if (m->surfaces[i].flags & SURF_DRAWTURB)
			{
				// subdivide for cleaner warps
				worldverts = GL_SubdivideSurface (m, m->surfaces + i, worldverts);
				continue;
			}

			// don't build polys for these surface types
			if (m->surfaces[i].flags & SURF_DRAWSKY) continue;

			// build regular polys
			worldverts = GL_BuildPolygonFromSurface (m, m->surfaces + i, worldverts);
		}
	}

	// allocate a contiguous array of polyverts for sending through the drawing routines
	polyverts = Hunk_Alloc (worldverts * sizeof (polyverts_t));

	// now we get to do it all over again!!!  this time, fill in the polyverts_t structs
	// the whole point of this is to get the polygon verts into a contiguous array that i can pass to the vertex
	// array drawing routines, thus avoiding extra memory transfers
	for (j = 1; j < MAX_MODELS; j++)
	{
		m = cl.model_precache[j];
		if (!m) break;

		// somebody in ID forgot this check, meaning that it will attempt to build polys for alias and sprite 
		// models too! (it won't of course, as they have no surfs, but it was still a mistake to forget it)
		if (m->type != mod_brush) continue;

		if (m->name[0] == '*') continue;

		for (i = 0; i < m->numsurfaces; i++)
		{
			// don't build polys for these surface types
			if (m->surfaces[i].flags & SURF_DRAWSKY) continue;

			// build polyverts
			GL_BuildPolyVerts (m->surfaces + i);
		}
	}
}


void R_DrawSbarBrushModel (model_t *mod)
{
	int i;
	int j;
	msurface_t *surf;
	float *v;
	float boxverts[3];
	int numverts = 0;

	float modHeight = mod->maxs[2] - mod->mins[2];
	float modWidth = mod->maxs[1] - mod->mins[1];
	float modDepth = mod->maxs[0] - mod->mins[0];

	qboolean lumablend = false;

	// get the first surf
	surf = &mod->surfaces[mod->firstmodelsurface];

	// darken it a bit
	glColor3f (0.7, 0.7, 0.7);

	// count the verts
	for (i = 0; i < mod->numsurfaces; i++, surf++)
	{
		// reconstruct the polygon
		GL_BuildPolygonFromSurface (mod, surf, 0);

		// accumulate the vertex counter
		numverts += surf->polys->numverts;
	}

	// build a fake polyverts for temporary usage
	polyverts = malloc (numverts * sizeof (polyverts_t));

	// reset the first surf
	surf = &mod->surfaces[mod->firstmodelsurface];
	numverts = 0;

	// these are normally 6 surfs with 4 verts per surf, but we can't assume that's a
	// general case...
	for (i = 0; i < mod->numsurfaces; i++, surf++)
	{
		// these shouldn't be present here but let's not mess things up anyway
		if (surf->flags & SURF_DRAWTURB) continue;
		if (surf->flags & SURF_DRAWSKY) continue;

		// link in the polygons
		surf->polys->vertoffset = numverts;
		numverts += surf->polys->numverts;
		GL_BuildPolyVerts (surf);

		if (surf->texinfo->texture->glt->lumatexnum) lumablend = true;

		glBindTexture (GL_TEXTURE_2D, surf->texinfo->texture->glt->texnum);

		glBegin (GL_TRIANGLE_FAN);

		// because we don't have polys set for anything here (so as not to mess up the vertex
		// array building) we must instead reconstruct the data from scratch)
		for (j = 0; j < surf->polys->numverts; j++)
		{
			// scale the verts so that it's sized correctly for the sbar
			boxverts[0] = ((surf->polys->basev[j][0] * 24) / modDepth) - 12;
			boxverts[1] = ((surf->polys->basev[j][1] * 24) / modWidth) - 12;
			boxverts[2] = ((surf->polys->basev[j][2] * 24) / modHeight) - 12;

			// draw it
			glTexCoord2fv (surf->polys->verts[j].st);
			glVertex3fv (boxverts);
		}

		glEnd ();
	}

	// back to normal colouring
	glColor3f (1, 1, 1);

	if (!lumablend)
	{
		// release our fake polyverts
		free (polyverts);

		return;
	}

	// blend mode for lumas
	glDepthMask (GL_FALSE);
	glEnable (GL_BLEND);
	glBlendFunc (GL_ONE, GL_ONE);

	// get the first surf again
	surf = &mod->surfaces[mod->firstmodelsurface];

	// luma luma luma
	for (i = 0; i < mod->numsurfaces; i++, surf++)
	{
		// these shouldn't be present here but let's not mess things up anyway
		if (surf->flags & SURF_DRAWTURB) continue;
		if (surf->flags & SURF_DRAWSKY) continue;

		if (!surf->texinfo->texture->glt->lumatexnum) continue;

		glBindTexture (GL_TEXTURE_2D, surf->texinfo->texture->glt->lumatexnum);

		glBegin (GL_TRIANGLE_FAN);

		// because we don't have polys set for anything here (so as not to mess up the vertex
		// array building) we must instead reconstruct the data from scratch)
		for (j = 0; j < surf->polys->numverts; j++)
		{
			// scale the verts so that it's sized correctly for the sbar
			boxverts[0] = ((surf->polys->basev[j][0] * 24) / modDepth) - 12;
			boxverts[1] = ((surf->polys->basev[j][1] * 24) / modWidth) - 12;
			boxverts[2] = ((surf->polys->basev[j][2] * 24) / modHeight) - 12;

			// draw it
			glTexCoord2fv (surf->polys->verts[j].st);
			glVertex3fv (boxverts);
		}

		glEnd ();
	}

	// restore the original blend mode
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);

	// release our fake polyverts
	free (polyverts);
}


void R_InitSurfs (void)
{
	int i;
	float j;
	float inc = (2.0 * M_PI) / (float) SINTABLE_SIZE;

	R_InitLight ();

	// set up the sin table for warps
	for (i = 0, j = 0; i < SINTABLE_SIZE; i++, j += inc)
		sintable[i] = sin (j) * 8.0;

	STBitMask = SINTABLE_SIZE - 1;

	STMult = (float) SINTABLE_SIZE / (2.0 * M_PI);

	// surface cvars
	Cvar_RegisterVariable (&r_bloom);
	Cvar_RegisterVariable (&r_bloomalpha);
	Cvar_RegisterVariable (&r_bloomradius);
}
