/*
===============================================================================
Copyright (C) 2006 Amarillo Hoss Trading Co., a wholly 
fictious affiliated partner of MH Enterprises, a wholly
fictitious conglomeration of affiliated partners.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

gl_verts.c
Vertex routines
===============================================================================
*/

#include "quakedef.h"


// moved the blend to CL_RelinkEntities
void R_RotateForEntity (entity_t *e)
{
	// go to a clean matrix because the transforms are going to be relative to the world matrix, not inclusive of it
	// this method saves on 64 bytes of bandwidth to the GPU per transformed model
	glLoadIdentity ();

	// run the transforms
	glTranslatef (e->origin[0], e->origin[1], e->origin[2]);

	// rotate y angle around z
	glRotatef (e->angles[1],  0, 0, 1);

	// store the matrix out for shadow drawing (saves us having to do all those transforms again).  
	// save it out *before* doing the rest of the rotation, cos that fucks with our shadow positioning
	if (e->model->type == mod_alias) glGetFloatv (GL_MODELVIEW_MATRIX, e->shadowmatrix);

	// now complete the rotation
	glRotatef (-e->angles[0],  0, 1, 0);
	glRotatef (e->angles[2],  1, 0, 0);

	// save out the full matrix
	glGetFloatv (GL_MODELVIEW_MATRIX, e->matrix);

	// restore the world matrix
	glLoadMatrixf (cl_entities[0].matrix);
}


/*
==================
R_TransformPolys

Transforms the "v" verts in a poly or chain of polys by the specified matrix, and
stores the result in the "tv" verts, which should then be used for rendering.

Works on all brush model surfaces (except skies, which are done very differently
anyway), and NOW on alias models too...!
==================
*/
void R_TransformPolys (glpoly_t *p, float *matrix)
{
	int i;
	float *v;
	float *tv;

	// do the full poly chain cos turb surfs can go through this too!!!
	for (; p; p = p->next)
	{
		// now transform the verts
		for (i = 0; i < p->numverts; i++)
		{
			// keep the loop clean
			v = p->basev[i];
			tv = p->verts[i].tv;

			// i used hexadecimal so it would line up nicely.  how sad...
			tv[0] = matrix[0x0] * v[0] + matrix[0x4] * v[1] + matrix[0x8] * v[2] + matrix[0xc];
			tv[1] = matrix[0x1] * v[0] + matrix[0x5] * v[1] + matrix[0x9] * v[2] + matrix[0xd];
			tv[2] = matrix[0x2] * v[0] + matrix[0x6] * v[1] + matrix[0xa] * v[2] + matrix[0xe];
		}
	}
}


// special case where the polys don't get a full transform but are only moved
void R_TranslatePolys (glpoly_t *p, vec3_t translate)
{
	int i;

	// do the full poly chain cos turb surfs can go through this too!!!
	for (; p; p = p->next)
	{
		// now translate the verts
		for (i = 0; i < p->numverts; i++)
		{
			p->verts[i].tv[0] = p->basev[i][0] + translate[0];
			p->verts[i].tv[1] = p->basev[i][1] + translate[1];
			p->verts[i].tv[2] = p->basev[i][2] + translate[2];
		}
	}
}


/*
========================================================================================================

	VERTEX ARRAYS

	Unified render for all the 3D stuff.

	This is what all the big mystery is about...  just 100 lines of code...

========================================================================================================
*/

static int VADrawFlags;

void GL_EnableVertexArrays (polyverts_t *VArrayVerts, int VAFlags)
{
	// sanity check
	if (VADrawFlags) GL_DisableVertexArrays ();

	// copy the flags across for global use
	VADrawFlags = VAFlags;

	// always have this, no matter what
	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (3, GL_FLOAT, sizeof (polyverts_t), (float *) VArrayVerts[0].tv);

	if (VADrawFlags & VA_DOT3)
	{
		// we get to reuse the verts for the tcs here
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (3, GL_FLOAT, sizeof (polyverts_t), (float *) VArrayVerts[0].tv);

		// regular st here
		glClientActiveTexture (GL_TEXTURE1);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), (float *) VArrayVerts[0].st);
	}
	else if (VADrawFlags & VA_STLMFB)
	{
		// world texture, lightmap and fullbright in a multitextured config
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), (float *) VArrayVerts[0].st);

		glClientActiveTexture (GL_TEXTURE1);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), (float *) VArrayVerts[0].lm);

		glClientActiveTexture (GL_TEXTURE2);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), (float *) VArrayVerts[0].st);
	}
	else if (VADrawFlags & VA_STLM)
	{
		// world texture and lightmap in a multitextured config
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), (float *) VArrayVerts[0].st);

		glClientActiveTexture (GL_TEXTURE1);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), (float *) VArrayVerts[0].lm);
	}
	else if (VADrawFlags & VA_STDUAL)
	{
		// 2 different texes using the same texcoords
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), (float *) VArrayVerts[0].st);

		glClientActiveTexture (GL_TEXTURE1);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), (float *) VArrayVerts[0].st);
	}
	else if (VADrawFlags & VA_ST)
	{
		// texture only
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), (float *) VArrayVerts[0].st);
	}
	else if (VADrawFlags & VA_LM)
	{
		// lightmap only
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), (float *) VArrayVerts[0].lm);
	}

	if (VADrawFlags & VA_RGBA)
	{
		// colour
		glEnableClientState (GL_COLOR_ARRAY);
		glColorPointer (4, GL_FLOAT, sizeof (polyverts_t), (float *) VArrayVerts[0].rgba);
	}
}


void GL_DisableVertexArrays (void)
{
	// sanity check
	if (!VADrawFlags) return;

	// disable them in the reverse order to which they came up
	// probably not necessary, but does no harm
	if (VADrawFlags & VA_RGBA)
		glDisableClientState (GL_COLOR_ARRAY);

	if (VADrawFlags & VA_STLMFB)
	{
		glClientActiveTexture (GL_TEXTURE2);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);

		glClientActiveTexture (GL_TEXTURE1);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);

		glClientActiveTexture (GL_TEXTURE0);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	}
	else if ((VADrawFlags & VA_STLM) || (VADrawFlags & VA_STDUAL) || ((VADrawFlags & VA_DOT3)))
	{
		glClientActiveTexture (GL_TEXTURE1);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);

		glClientActiveTexture (GL_TEXTURE0);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	}
	else if (VADrawFlags & VA_ST)
	{
		glClientActiveTexture (GL_TEXTURE0);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	}

	glDisableClientState (GL_VERTEX_ARRAY);

	VADrawFlags = 0;
}


void GL_DrawVertexArrays (glpoly_t *poly)
{
	for (; poly; poly = poly->next)
		glDrawArrays (poly->PRIMITIVE_TYPE, poly->vertoffset, poly->numverts);
}
